@@ -154,6 +154,8 @@ class GetContext(WSGIContext):
 
         con_resp = con_req.get_response(self.dlo.app)
         if not is_success(con_resp.status_int):
+            if req.method == 'HEAD':
+                con_resp.body = ''
             return con_resp, None
         with closing_if_possible(con_resp.app_iter):
             return None, json.loads(''.join(con_resp.app_iter))
@@ -281,6 +283,8 @@ class GetContext(WSGIContext):
                 actual_content_length = None
                 content_length_for_swob_range = None
                 req.range = None
+        else:
+            req.range = None
 
         response_headers = [
             (h, v) for h, v in response_headers
