@@ -34,8 +34,9 @@ from swift.common.request_helpers import get_param, \
 from swift.common.utils import get_logger, hash_path, public, \
     Timestamp, storage_directory, validate_sync_to, \
     config_true_value, timing_stats, replication, \
-    override_bytes_from_content_type, get_log_line, ShardRange, list_from_csv
-
+    override_bytes_from_content_type, get_log_line, \
+    config_fallocate_value, fs_has_free_space, list_from_csv, \
+    ShardRange
 from swift.common.constraints import valid_timestamp, check_utf8, check_drive
 from swift.common import constraints
 from swift.common.bufferedhttp import http_connect
@@ -124,6 +125,8 @@ class ContainerController(BaseStorageServer):
         self.sync_store = ContainerSyncStore(self.root,
                                              self.logger,
                                              self.mount_check)
+        self.fallocate_reserve, self.fallocate_is_percent = \
+            config_fallocate_value(conf.get('fallocate_reserve', '1%'))
 
     def _get_container_broker(self, drive, part, account, container, **kwargs):
         """
@@ -298,6 +301,11 @@ class ContainerController(BaseStorageServer):
         req.environ['swift.leave_relative_location'] = True
         return HTTPMovedPermanently(headers=headers, request=req)
 
+    def check_free_space(self, drive):
+        drive_root = os.path.join(self.root, drive)
+        return fs_has_free_space(
+            drive_root, self.fallocate_reserve, self.fallocate_is_percent)
+
     @public
     @timing_stats()
     def DELETE(self, req):
@@ -305,7 +313,9 @@ class ContainerController(BaseStorageServer):
         drive, part, account, container, obj = split_and_validate_path(
             req, 4, 5, True)
         req_timestamp = valid_timestamp(req)
-        if not check_drive(self.root, drive, self.mount_check):
+        try:
+            check_drive(self.root, drive, self.mount_check)
+        except ValueError:
             return HTTPInsufficientStorage(drive=drive, request=req)
         # policy index is only relevant for delete_obj (and transitively for
         # auto create accounts)
@@ -432,7 +442,11 @@ class ContainerController(BaseStorageServer):
                 self.realms_conf)
             if err:
                 return HTTPBadRequest(err)
-        if not check_drive(self.root, drive, self.mount_check):
+        try:
+            check_drive(self.root, drive, self.mount_check)
+        except ValueError:
+            return HTTPInsufficientStorage(drive=drive, request=req)
+        if not self.check_free_space(drive):
             return HTTPInsufficientStorage(drive=drive, request=req)
         requested_policy_index = self.get_and_validate_policy_index(req)
         broker = self._get_container_broker(drive, part, account, container)
@@ -502,7 +516,9 @@ class ContainerController(BaseStorageServer):
         drive, part, account, container, obj = split_and_validate_path(
             req, 4, 5, True)
         out_content_type = listing_formats.get_listing_content_type(req)
-        if not check_drive(self.root, drive, self.mount_check):
+        try:
+            check_drive(self.root, drive, self.mount_check)
+        except ValueError:
             return HTTPInsufficientStorage(drive=drive, request=req)
         broker = self._get_container_broker(drive, part, account, container,
                                             pending_timeout=0.1,
@@ -549,7 +565,71 @@ class ContainerController(BaseStorageServer):
     @public
     @timing_stats()
     def GET(self, req):
-        """Handle HTTP GET request."""
+        """
+        Handle HTTP GET request.
+
+        The body of the response to a successful GET request contains a listing
+        of either objects or shard ranges. The exact content of the listing is
+        determined by a combination of request headers and query string
+        parameters, as follows:
+
+        * The type of the listing is determined by the
+          ``X-Backend-Record-Type`` header. If this header has value ``shard``
+          then the response body will be a list of shard ranges; if this header
+          has value ``auto``, and the container state is ``sharding`` or
+          ``sharded``, then the listing will be a list of shard ranges;
+          otherwise the response body will be a list of objects.
+
+        * Both shard range and object listings may be constrained to a name
+          range by the ``marker`` and ``end_marker`` query string parameters.
+          Object listings will only contain objects whose names are greater
+          than any ``marker`` value and less than any ``end_marker`` value.
+          Shard range listings will only contain shard ranges whose namespace
+          is greater than or includes any ``marker`` value and is less than or
+          includes any ``end_marker`` value.
+
+        * Shard range listings may also be constrained by an ``includes`` query
+          string parameter. If this parameter is present the listing will only
+          contain shard ranges whose namespace includes the value of the
+          parameter; any ``marker`` or ``end_marker`` parameters are ignored
+
+        * The length of an object listing may be constrained by the ``limit``
+          parameter. Object listings may also be constrained by ``prefix``,
+          ``delimiter`` and ``path`` query string parameters.
+
+        * Shard range listings will include deleted shard ranges if and only if
+          the ``X-Backend-Include-Deleted`` header value is one of
+          :attr:`swift.common.utils.TRUE_VALUES`. Object listings never
+          include deleted objects.
+
+        * Shard range listings may be constrained to include only shard ranges
+          whose state is specified by a query string ``states`` parameter. If
+          present, the ``states`` parameter should be a comma separated list of
+          either the string or integer representation of
+          :data:`~swift.common.utils.ShardRange.STATES`.
+
+          Two alias values may be used in a ``states`` parameter value:
+          ``listing`` will cause the listing to include all shard ranges in a
+          state suitable for contributing to an object listing; ``updating``
+          will cause the listing to include all shard ranges in a state
+          suitable to accept an object update.
+
+          If either of these aliases is used then the shard range listing will
+          if necessary be extended with a synthesised 'filler' range in order
+          to satisfy the requested name range when insufficient actual shard
+          ranges are found. Any 'filler' shard range will cover the otherwise
+          uncovered tail of the requested name range and will point back to the
+          same container.
+
+        * Listings are not normally returned from a deleted container. However,
+          the ``X-Backend-Override-Deleted`` header may be used with a value in
+          :attr:`swift.common.utils.TRUE_VALUES` to force a shard range
+          listing to be returned from a deleted container whose DB file still
+          exists.
+
+        :param req: an instance of :class:`swift.common.swob.Request`
+        :returns: an instance of :class:`swift.common.swob.Response`
+        """
         drive, part, account, container, obj = split_and_validate_path(
             req, 4, 5, True)
         path = get_param(req, 'path')
@@ -571,7 +651,9 @@ class ContainerController(BaseStorageServer):
                     body='Maximum limit is %d'
                     % constraints.CONTAINER_LISTING_LIMIT)
         out_content_type = listing_formats.get_listing_content_type(req)
-        if not check_drive(self.root, drive, self.mount_check):
+        try:
+            check_drive(self.root, drive, self.mount_check)
+        except ValueError:
             return HTTPInsufficientStorage(drive=drive, request=req)
         broker = self._get_container_broker(drive, part, account, container,
                                             pending_timeout=0.1,
@@ -650,7 +732,11 @@ class ContainerController(BaseStorageServer):
         """
         post_args = split_and_validate_path(req, 3)
         drive, partition, hash = post_args
-        if not check_drive(self.root, drive, self.mount_check):
+        try:
+            check_drive(self.root, drive, self.mount_check)
+        except ValueError:
+            return HTTPInsufficientStorage(drive=drive, request=req)
+        if not self.check_free_space(drive):
             return HTTPInsufficientStorage(drive=drive, request=req)
         try:
             args = json.load(req.environ['wsgi.input'])
@@ -672,7 +758,11 @@ class ContainerController(BaseStorageServer):
                 self.realms_conf)
             if err:
                 return HTTPBadRequest(err)
-        if not check_drive(self.root, drive, self.mount_check):
+        try:
+            check_drive(self.root, drive, self.mount_check)
+        except ValueError:
+            return HTTPInsufficientStorage(drive=drive, request=req)
+        if not self.check_free_space(drive):
             return HTTPInsufficientStorage(drive=drive, request=req)
         broker = self._get_container_broker(drive, part, account, container)
         if broker.is_deleted():
