@@ -17,6 +17,7 @@ import hashlib
 import json
 import os
 import random
+import sys
 import socket
 import time
 
@@ -104,6 +105,91 @@ def listing_items(method):
             items = []
 
 
+def putrequest(self, method, url, skip_host=False, skip_accept_encoding=False):
+    '''Send a request to the server.
+
+    This is mostly a regurgitation of CPython's HTTPConnection.putrequest,
+    but fixed up so we can still send arbitrary bytes in the request line
+    on py3. See also: https://bugs.python.org/issue36274
+
+    To use, swap out a HTTP(S)Connection's putrequest with something like::
+
+       conn.putrequest = putrequest.__get__(conn)
+
+    :param method: specifies an HTTP request method, e.g. 'GET'.
+    :param url: specifies the object being requested, e.g. '/index.html'.
+    :param skip_host: if True does not add automatically a 'Host:' header
+    :param skip_accept_encoding: if True does not add automatically an
+       'Accept-Encoding:' header
+    '''
+    # (Mostly) inline the HTTPConnection implementation; just fix it
+    # so we can send non-ascii request lines. For comparison, see
+    # https://github.com/python/cpython/blob/v2.7.16/Lib/httplib.py#L888-L1003
+    # and https://github.com/python/cpython/blob/v3.7.2/
+    # Lib/http/client.py#L1061-L1183
+    if self._HTTPConnection__response \
+            and self._HTTPConnection__response.isclosed():
+        self._HTTPConnection__response = None
+
+    if self._HTTPConnection__state == http_client._CS_IDLE:
+        self._HTTPConnection__state = http_client._CS_REQ_STARTED
+    else:
+        raise http_client.CannotSendRequest(self._HTTPConnection__state)
+
+    self._method = method
+    if not url:
+        url = '/'
+    self._path = url
+    request = '%s %s %s' % (method, url, self._http_vsn_str)
+    if not isinstance(request, bytes):
+        # This choice of encoding is the whole reason we copy/paste from
+        # cpython. When making backend requests, it should never be
+        # necessary; however, we have some functional tests that want
+        # to send non-ascii bytes.
+        # TODO: when https://bugs.python.org/issue36274 is resolved, make
+        # sure we fix up our API to match whatever upstream chooses to do
+        self._output(request.encode('latin1'))
+    else:
+        self._output(request)
+
+    if self._http_vsn == 11:
+        if not skip_host:
+            netloc = ''
+            if url.startswith('http'):
+                nil, netloc, nil, nil, nil = urllib.parse.urlsplit(url)
+
+            if netloc:
+                try:
+                    netloc_enc = netloc.encode("ascii")
+                except UnicodeEncodeError:
+                    netloc_enc = netloc.encode("idna")
+                self.putheader('Host', netloc_enc)
+            else:
+                if self._tunnel_host:
+                    host = self._tunnel_host
+                    port = self._tunnel_port
+                else:
+                    host = self.host
+                    port = self.port
+
+                try:
+                    host_enc = host.encode("ascii")
+                except UnicodeEncodeError:
+                    host_enc = host.encode("idna")
+
+                if host.find(':') >= 0:
+                    host_enc = b'[' + host_enc + b']'
+
+                if port == self.default_port:
+                    self.putheader('Host', host_enc)
+                else:
+                    host_enc = host_enc.decode("ascii")
+                    self.putheader('Host', "%s:%s" % (host_enc, port))
+
+        if not skip_accept_encoding:
+            self.putheader('Accept-Encoding', 'identity')
+
+
 class Connection(object):
     def __init__(self, config):
         for key in 'auth_host auth_port auth_ssl username password'.split():
@@ -123,21 +209,47 @@ class Connection(object):
         self.password = config['password']
 
         self.storage_netloc = None
-        self.storage_url = None
-
+        self.storage_path = None
         self.conn_class = None
+        self.connection = None  # until you call .http_connect()
+
+    @property
+    def storage_url(self):
+        return '%s://%s/%s' % (self.storage_scheme, self.storage_netloc,
+                               self.storage_path)
+
+    @storage_url.setter
+    def storage_url(self, value):
+        url = urllib.parse.urlparse(value)
+
+        if url.scheme == 'http':
+            self.conn_class = http_client.HTTPConnection
+        elif url.scheme == 'https':
+            self.conn_class = http_client.HTTPSConnection
+        else:
+            raise ValueError('unexpected protocol %s' % (url.scheme))
+
+        self.storage_netloc = url.netloc
+        # Make sure storage_path is a string and not unicode, since
+        # keystoneclient (called by swiftclient) returns them in
+        # unicode and this would cause troubles when doing
+        # no_safe_quote query.
+        x = url.path.split('/')
+        self.storage_path = str('/%s/%s' % (x[1], x[2]))
+        self.account_name = str(x[2])
+
+    @property
+    def storage_scheme(self):
+        if self.conn_class is None:
+            return None
+        if issubclass(self.conn_class, http_client.HTTPSConnection):
+            return 'https'
+        return 'http'
 
     def get_account(self):
         return Account(self, self.account)
 
-    def authenticate(self, clone_conn=None):
-        if clone_conn:
-            self.conn_class = clone_conn.conn_class
-            self.storage_netloc = clone_conn.storage_netloc
-            self.storage_url = clone_conn.storage_url
-            self.storage_token = clone_conn.storage_token
-            return
-
+    def authenticate(self):
         if self.auth_version == "1":
             auth_path = '%sv1.0' % (self.auth_prefix)
             if self.account:
@@ -151,6 +263,16 @@ class Connection(object):
         auth_netloc = "%s:%d" % (self.auth_host, self.auth_port)
         auth_url = auth_scheme + auth_netloc + auth_path
 
+        if self.insecure:
+            try:
+                import requests
+                from requests.packages.urllib3.exceptions import \
+                    InsecureRequestWarning
+            except ImportError:
+                pass
+            else:
+                requests.packages.urllib3.disable_warnings(
+                    InsecureRequestWarning)
         authargs = dict(snet=False, tenant_name=self.account,
                         auth_version=self.auth_version, os_options={},
                         insecure=self.insecure)
@@ -160,23 +282,7 @@ class Connection(object):
         if not (storage_url and storage_token):
             raise AuthenticationFailed()
 
-        url = urllib.parse.urlparse(storage_url)
-
-        if url.scheme == 'http':
-            self.conn_class = http_client.HTTPConnection
-        elif url.scheme == 'https':
-            self.conn_class = http_client.HTTPSConnection
-        else:
-            raise ValueError('unexpected protocol %s' % (url.scheme))
-
-        self.storage_netloc = url.netloc
-        # Make sure storage_url is a string and not unicode, since
-        # keystoneclient (called by swiftclient) returns them in
-        # unicode and this would cause troubles when doing
-        # no_safe_quote query.
-        x = url.path.split('/')
-        self.storage_url = str('/%s/%s' % (x[1], x[2]))
-        self.account_name = str(x[2])
+        self.storage_url = storage_url
         self.auth_user = auth_user
         # With v2 keystone, storage_token is unicode.
         # We want it to be string otherwise this would cause
@@ -186,7 +292,7 @@ class Connection(object):
         self.user_acl = '%s:%s' % (self.account, self.username)
 
         self.http_connect()
-        return self.storage_url, self.storage_token
+        return self.storage_path, self.storage_token
 
     def cluster_info(self):
         """
@@ -201,7 +307,15 @@ class Connection(object):
         return json.loads(self.response.read())
 
     def http_connect(self):
-        self.connection = self.conn_class(self.storage_netloc)
+        if self.storage_scheme == 'https' and \
+                self.insecure and sys.version_info >= (2, 7, 9):
+            import ssl
+            self.connection = self.conn_class(
+                self.storage_netloc,
+                context=ssl._create_unverified_context())
+        else:
+            self.connection = self.conn_class(self.storage_netloc)
+        self.connection.putrequest = putrequest.__get__(self.connection)
 
     def make_path(self, path=None, cfg=None):
         if path is None:
@@ -210,16 +324,16 @@ class Connection(object):
             cfg = {}
 
         if cfg.get('version_only_path'):
-            return '/' + self.storage_url.split('/')[1]
+            return '/' + self.storage_path.split('/')[1]
 
         if path:
             quote = urllib.parse.quote
             if cfg.get('no_quote') or cfg.get('no_path_quote'):
                 quote = lambda x: x
-            return '%s/%s' % (self.storage_url,
+            return '%s/%s' % (self.storage_path,
                               '/'.join([quote(i) for i in path]))
         else:
-            return self.storage_url
+            return self.storage_path
 
     def make_headers(self, hdrs, cfg=None):
         if cfg is None:
@@ -328,7 +442,6 @@ class Connection(object):
                           for (x, y) in parms.items()]
             path = '%s?%s' % (path, '&'.join(query_args))
 
-        self.connection = self.conn_class(self.storage_netloc)
         self.connection.putrequest('PUT', path)
         for key, value in headers.items():
             self.connection.putheader(key, value)
@@ -804,7 +917,7 @@ class File(Base):
 
         header_fields = self.header_fields(fields,
                                            optional_fields=optional_fields)
-        header_fields['etag'] = header_fields['etag'].strip('"')
+        header_fields['etag'] = header_fields['etag']
         return header_fields
 
     def initialize(self, hdrs=None, parms=None):
@@ -829,7 +942,7 @@ class File(Base):
             if hdr[0].lower().startswith('x-object-meta-'):
                 self.metadata[hdr[0][14:]] = hdr[1]
             if hdr[0].lower() == 'etag':
-                self.etag = hdr[1].strip('"')
+                self.etag = hdr[1]
             if hdr[0].lower() == 'content-length':
                 self.size = int(hdr[1])
             if hdr[0].lower() == 'last-modified':
