@@ -31,6 +31,7 @@ from swift.cli.info import (print_db_info_metadata, print_ring_locations,
                             parse_get_node_args)
 from swift.account.server import AccountController
 from swift.container.server import ContainerController
+from swift.container.backend import UNSHARDED, SHARDED
 from swift.obj.diskfile import write_metadata
 
 
@@ -42,8 +43,8 @@ class TestCliInfoBase(unittest.TestCase):
     def setUp(self):
         skip_if_no_xattrs()
         self.orig_hp = utils.HASH_PATH_PREFIX, utils.HASH_PATH_SUFFIX
-        utils.HASH_PATH_PREFIX = 'info'
-        utils.HASH_PATH_SUFFIX = 'info'
+        utils.HASH_PATH_PREFIX = b'info'
+        utils.HASH_PATH_SUFFIX = b'info'
         self.testdir = os.path.join(mkdtemp(), 'tmp_test_cli_info')
         utils.mkdirs(self.testdir)
         rmtree(self.testdir)
@@ -103,17 +104,18 @@ class TestCliInfo(TestCliInfoBase):
         self.assertRaisesMessage(ValueError, 'Info is incomplete',
                                  print_db_info_metadata, 'container', {}, {})
 
-        info = dict(
-            account='acct',
-            created_at=100.1,
-            put_timestamp=106.3,
-            delete_timestamp=107.9,
-            status_changed_at=108.3,
-            container_count='3',
-            object_count='20',
-            bytes_used='42')
-        info['hash'] = 'abaddeadbeefcafe'
-        info['id'] = 'abadf100d0ddba11'
+        info = {
+            'account': 'acct',
+            'created_at': 100.1,
+            'put_timestamp': 106.3,
+            'delete_timestamp': 107.9,
+            'status_changed_at': 108.3,
+            'container_count': '3',
+            'object_count': '20',
+            'bytes_used': '42',
+            'hash': 'abaddeadbeefcafe',
+            'id': 'abadf100d0ddba11',
+        }
         md = {'x-account-meta-mydata': ('swift', '0000000000.00000'),
               'x-other-something': ('boo', '0000000000.00000')}
         out = StringIO()
@@ -134,7 +136,7 @@ Metadata:
   UUID: abadf100d0ddba11
   X-Other-Something: boo
 No system metadata found in db file
-  User Metadata: {'mydata': 'swift'}'''
+  User Metadata: {'x-account-meta-mydata': 'swift'}'''
 
         self.assertEqual(sorted(out.getvalue().strip().split('\n')),
                          sorted(exp_out.split('\n')))
@@ -154,13 +156,15 @@ No system metadata found in db file
             reported_object_count='20',
             reported_bytes_used='42',
             x_container_foo='bar',
-            x_container_bar='goo')
+            x_container_bar='goo',
+            db_state=UNSHARDED,
+            is_root=True)
         info['hash'] = 'abaddeadbeefcafe'
         info['id'] = 'abadf100d0ddba11'
         md = {'x-container-sysmeta-mydata': ('swift', '0000000000.00000')}
         out = StringIO()
         with mock.patch('sys.stdout', out):
-            print_db_info_metadata('container', info, md)
+            print_db_info_metadata('container', info, md, True)
         exp_out = '''Path: /acct/cont
   Account: acct
   Container: cont
@@ -182,10 +186,88 @@ Metadata:
   X-Container-Bar: goo
   X-Container-Foo: bar
   System Metadata: {'mydata': 'swift'}
-No user metadata found in db file''' % POLICIES[0].name
+No user metadata found in db file
+Sharding Metadata:
+  Type: root
+  State: unsharded''' % POLICIES[0].name
         self.assertEqual(sorted(out.getvalue().strip().split('\n')),
                          sorted(exp_out.split('\n')))
 
+    def test_print_db_info_metadata_with_shard_ranges(self):
+
+        shard_ranges = [utils.ShardRange(
+            name='.sharded_a/shard_range_%s' % i,
+            timestamp=utils.Timestamp(i), lower='%da' % i,
+            upper='%dz' % i, object_count=i, bytes_used=i,
+            meta_timestamp=utils.Timestamp(i)) for i in range(1, 4)]
+        shard_ranges[0].state = utils.ShardRange.CLEAVED
+        shard_ranges[1].state = utils.ShardRange.CREATED
+
+        info = dict(
+            account='acct',
+            container='cont',
+            storage_policy_index=0,
+            created_at='0000000100.10000',
+            put_timestamp='0000000106.30000',
+            delete_timestamp='0000000107.90000',
+            status_changed_at='0000000108.30000',
+            object_count='20',
+            bytes_used='42',
+            reported_put_timestamp='0000010106.30000',
+            reported_delete_timestamp='0000010107.90000',
+            reported_object_count='20',
+            reported_bytes_used='42',
+            db_state=SHARDED,
+            is_root=True,
+            shard_ranges=shard_ranges)
+        info['hash'] = 'abaddeadbeefcafe'
+        info['id'] = 'abadf100d0ddba11'
+        out = StringIO()
+        with mock.patch('sys.stdout', out):
+            print_db_info_metadata('container', info, {})
+        exp_out = '''Path: /acct/cont
+  Account: acct
+  Container: cont
+  Container Hash: d49d0ecbb53be1fcc49624f2f7c7ccae
+Metadata:
+  Created at: 1970-01-01T00:01:40.100000 (0000000100.10000)
+  Put Timestamp: 1970-01-01T00:01:46.300000 (0000000106.30000)
+  Delete Timestamp: 1970-01-01T00:01:47.900000 (0000000107.90000)
+  Status Timestamp: 1970-01-01T00:01:48.300000 (0000000108.30000)
+  Object Count: 20
+  Bytes Used: 42
+  Storage Policy: %s (0)
+  Reported Put Timestamp: 1970-01-01T02:48:26.300000 (0000010106.30000)
+  Reported Delete Timestamp: 1970-01-01T02:48:27.900000 (0000010107.90000)
+  Reported Object Count: 20
+  Reported Bytes Used: 42
+  Chexor: abaddeadbeefcafe
+  UUID: abadf100d0ddba11
+No system metadata found in db file
+No user metadata found in db file
+Sharding Metadata:
+  Type: root
+  State: sharded
+Shard Ranges (3):
+  Name: .sharded_a/shard_range_1
+    lower: '1a', upper: '1z'
+    Object Count: 1, Bytes Used: 1, State: cleaved (30)
+    Created at: 1970-01-01T00:00:01.000000 (0000000001.00000)
+    Meta Timestamp: 1970-01-01T00:00:01.000000 (0000000001.00000)
+  Name: .sharded_a/shard_range_2
+    lower: '2a', upper: '2z'
+    Object Count: 2, Bytes Used: 2, State: created (20)
+    Created at: 1970-01-01T00:00:02.000000 (0000000002.00000)
+    Meta Timestamp: 1970-01-01T00:00:02.000000 (0000000002.00000)
+  Name: .sharded_a/shard_range_3
+    lower: '3a', upper: '3z'
+    Object Count: 3, Bytes Used: 3, State: found (10)
+    Created at: 1970-01-01T00:00:03.000000 (0000000003.00000)
+    Meta Timestamp: 1970-01-01T00:00:03.000000 (0000000003.00000)''' %\
+                  POLICIES[0].name
+        self.assertEqual(sorted(out.getvalue().strip().split('\n')),
+                         sorted(exp_out.strip().split('\n')))
+
     def test_print_ring_locations_invalid_args(self):
         self.assertRaises(ValueError, print_ring_locations,
                           None, 'dir', 'acct')
@@ -423,14 +505,8 @@ No user metadata found in db file''' % POLICIES[0].name
                                    '1', 'b47',
                                    'dc5be2aa4347a22a0fee6bc7de505b47',
                                    'dc5be2aa4347a22a0fee6bc7de505b47.db')
-            try:
-                print_info('account', db_file, swift_dir=self.testdir)
-            except Exception:
-                exp_raised = True
-        if exp_raised:
-            self.fail("Unexpected exception raised")
-        else:
-            self.assertGreater(len(out.getvalue().strip()), 800)
+            print_info('account', db_file, swift_dir=self.testdir)
+        self.assertGreater(len(out.getvalue().strip()), 800)
 
         controller = ContainerController(
             {'devices': self.testdir, 'mount_check': 'false'})
@@ -875,7 +951,7 @@ class TestPrintObj(TestCliInfoBase):
         self.assertRaises(InfoSystemExit, print_obj, datafile)
 
         with open(datafile, 'wb') as fp:
-            fp.write('1234')
+            fp.write(b'1234')
 
         out = StringIO()
         with mock.patch('sys.stdout', out):
@@ -1129,7 +1205,7 @@ Other Metadata:
         })
         out = StringIO()
         with mock.patch('sys.stdout', out):
-            print_obj_metadata(metadata)
+            print_obj_metadata(metadata, True)
         exp_out = '''Path: /AUTH_admin/c/dummy
   Account: AUTH_admin
   Container: c
@@ -1138,8 +1214,8 @@ Other Metadata:
 Content-Type: application/octet-stream
 Timestamp: 1970-01-01T00:01:46.300000 (%s)
 System Metadata:
-  X-Object-Sysmeta-Mtime: 107.3
-  X-Object-Sysmeta-Name: Obj name
+  Mtime: 107.3
+  Name: Obj name
 Transient System Metadata:
   No metadata found
 User Metadata:
@@ -1209,7 +1285,7 @@ Other Metadata:
         del metadata['name']
         out = StringIO()
         with mock.patch('sys.stdout', out):
-            print_obj_metadata(metadata)
+            print_obj_metadata(metadata, True)
         exp_out = '''Path: Not found in metadata
 Content-Type: application/octet-stream
 Timestamp: 1970-01-01T00:01:46.300000 (%s)
@@ -1218,7 +1294,7 @@ System Metadata:
 Transient System Metadata:
   No metadata found
 User Metadata:
-  X-Object-Meta-Mtime: 107.3
+  Mtime: 107.3
 Other Metadata:
   No metadata found''' % (
             utils.Timestamp(106.3).internal)
@@ -1253,7 +1329,7 @@ Other Metadata:
         del metadata['X-Timestamp']
         out = StringIO()
         with mock.patch('sys.stdout', out):
-            print_obj_metadata(metadata)
+            print_obj_metadata(metadata, True)
         exp_out = '''Path: /AUTH_admin/c/dummy
   Account: AUTH_admin
   Container: c
@@ -1266,7 +1342,7 @@ System Metadata:
 Transient System Metadata:
   No metadata found
 User Metadata:
-  X-Object-Meta-Mtime: 107.3
+  Mtime: 107.3
 Other Metadata:
   No metadata found'''
 
@@ -1300,6 +1376,34 @@ Other Metadata:
 
         self.assertEqual(out.getvalue().strip(), exp_out)
 
+        metadata = get_metadata({
+            'X-Object-Meta-Mtime': '107.3',
+            'X-Object-Sysmeta-Mtime': '106.3',
+            'X-Object-Transient-Sysmeta-Mtime': '105.3',
+            'X-Object-Mtime': '104.3',
+        })
+        out = StringIO()
+        with mock.patch('sys.stdout', out):
+            print_obj_metadata(metadata, True)
+        exp_out = '''Path: /AUTH_admin/c/dummy
+  Account: AUTH_admin
+  Container: c
+  Object: dummy
+  Object hash: 128fdf98bddd1b1e8695f4340e67a67a
+Content-Type: application/octet-stream
+Timestamp: 1970-01-01T00:01:46.300000 (%s)
+System Metadata:
+  Mtime: 106.3
+Transient System Metadata:
+  Mtime: 105.3
+User Metadata:
+  Mtime: 107.3
+Other Metadata:
+  X-Object-Mtime: 104.3''' % (
+            utils.Timestamp(106.3).internal)
+
+        self.assertEqual(out.getvalue().strip(), exp_out)
+
 
 class TestPrintObjWeirdPath(TestPrintObjFullMeta):
     def setUp(self):
