@@ -13,11 +13,11 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import json
 import unittest
 from datetime import datetime
 from hashlib import md5
 
-from six.moves import urllib
 from swift.common import swob
 from swift.common.swob import Request
 
@@ -65,8 +65,15 @@ class TestS3ApiMultiDelete(S3ApiTestCase):
                             swob.HTTPNoContent, {}, None)
         self.swift.register('DELETE', '/v1/AUTH_test/bucket/Key2',
                             swob.HTTPNotFound, {}, None)
+        slo_delete_resp = {
+            'Number Not Found': 0,
+            'Response Status': '200 OK',
+            'Errors': [],
+            'Response Body': '',
+            'Number Deleted': 8
+        }
         self.swift.register('DELETE', '/v1/AUTH_test/bucket/Key3',
-                            swob.HTTPOk, {}, None)
+                            swob.HTTPOk, {}, json.dumps(slo_delete_resp))
 
         elem = Element('Delete')
         for key in ['Key1', 'Key2', 'Key3']:
@@ -87,11 +94,14 @@ class TestS3ApiMultiDelete(S3ApiTestCase):
 
         elem = fromstring(body)
         self.assertEqual(len(elem.findall('Deleted')), 3)
-        _, path, _ = self.swift.calls_with_headers[-1]
-        path, query_string = path.split('?', 1)
-        self.assertEqual(path, '/v1/AUTH_test/bucket/Key3')
-        query = dict(urllib.parse.parse_qsl(query_string))
-        self.assertEqual(query['multipart-manifest'], 'delete')
+        self.assertEqual(self.swift.calls, [
+            ('HEAD', '/v1/AUTH_test/bucket'),
+            ('HEAD', '/v1/AUTH_test/bucket/Key1'),
+            ('DELETE', '/v1/AUTH_test/bucket/Key1'),
+            ('HEAD', '/v1/AUTH_test/bucket/Key2'),
+            ('HEAD', '/v1/AUTH_test/bucket/Key3'),
+            ('DELETE', '/v1/AUTH_test/bucket/Key3?multipart-manifest=delete'),
+        ])
 
     @s3acl
     def test_object_multi_DELETE_quiet(self):
@@ -177,12 +187,37 @@ class TestS3ApiMultiDelete(S3ApiTestCase):
         status, headers, body = self.call_s3api(req)
         self.assertEqual(self._get_error_code(body), 'InvalidRequest')
 
+    @s3acl
+    def test_object_multi_DELETE_lots_of_keys(self):
+        elem = Element('Delete')
+        for i in range(self.conf.max_multi_delete_objects):
+            name = 'x' * 1000 + str(i)
+            self.swift.register('HEAD', '/v1/AUTH_test/bucket/%s' % name,
+                                swob.HTTPNotFound, {}, None)
+            obj = SubElement(elem, 'Object')
+            SubElement(obj, 'Key').text = name
+        body = tostring(elem, use_s3ns=False)
+        content_md5 = md5(body).digest().encode('base64').strip()
+
+        req = Request.blank('/bucket?delete',
+                            environ={'REQUEST_METHOD': 'POST'},
+                            headers={'Authorization': 'AWS test:tester:hmac',
+                                     'Date': self.get_date_header(),
+                                     'Content-MD5': content_md5},
+                            body=body)
+        status, headers, body = self.call_s3api(req)
+        self.assertEqual('200 OK', status)
+
+        elem = fromstring(body)
+        self.assertEqual(len(elem.findall('Deleted')),
+                         self.conf.max_multi_delete_objects)
+
     @s3acl
     def test_object_multi_DELETE_too_many_keys(self):
         elem = Element('Delete')
         for i in range(self.conf.max_multi_delete_objects + 1):
             obj = SubElement(elem, 'Object')
-            SubElement(obj, 'Key').text = str(i)
+            SubElement(obj, 'Key').text = 'x' * 1000 + str(i)
         body = tostring(elem, use_s3ns=False)
         content_md5 = md5(body).digest().encode('base64').strip()
 
